// define optional gl functions
// this file gets included multiple times, with different
// definitions for GL_FUNC

GL_FUNC (PFNGLBLENDFUNCSEPARATEPROC, BlendFuncSeparate)
GL_FUNC (PFNGLBLENDFUNCSEPARATEEXTPROC, BlendFuncSeparateEXT)
GL_FUNC (PFNGLCONVOLUTIONPARAMETERFEXTPROC, ConvolutionParameterf)
GL_FUNC (PFNGLCONVOLUTIONFILTER2DEXTPROC, ConvolutionFilter2D)
GL_FUNC (PFNGLSEPARABLEFILTER2DEXTPROC, SeparableFilter2D)
